/*********************************************************************************************************
Class Category
Files: category.h, category.cpp
Authors: Muhammad Zulfigar and Thalia Lewis
JPC36 Wx-View
Functions definitions for class Category
Last updated: August 9, 2000

Members 								Purpose
string name 						string that holds the name of the category
vector<string> files 				vector that holds paths to files belonging to the Category
vector<Category> subcategories 		vector that holds other Category objects belonging to the Category
wxTreeItemId category_tree_id; 		stores the id that associates this Category object with a certain item on a treeCtrl					
	
**********************************************************************************************************/

#include "category.h"

#include <string>
#include <vector>
using namespace std;


/********************************************************************************************
	Function: Category::add_file
	Receives: const string & file_name, const wxTreeItemId &selected_id
	Returns: void
	Description: Adds a filename to the files vector of the category object associated
				with the given wxTreeItemId.
*********************************************************************************************/
void Category::add_file(const string &file_name, const wxTreeItemId &selected_id)
{
	Category * node = find_id(selected_id);
	
	if (node != NULL)
	{
		node->files.push_back(file_name);
	}
}

/********************************************************************************************
	Function: Category::add_subcategory
	Receives: const string & new_subcategory_name, const wxTreeItemId &selected_id
	Returns: void
	Description: Adds a subcategory to the subcategories vector of the category object associated
				with the given wxTreeItemId.
*********************************************************************************************/
void Category::add_subcategory(const string &new_subcategory_name, const wxTreeItemId & selected_id)
{
	Category * node = find_id (selected_id);
	
	if (node != NULL)
	{
		node->subcategories.push_back(new_subcategory_name);
	}	 	
}

/********************************************************************************************
	Function: Category::fill_categories
	Receives: const vector<string> & file_name, int & current_line
	Returns: void
	Description: Recursively adds files and subcategories to category objects, representing
				the contents of the "categories.txt" file.Whether the item on the current line
				is added as a "sibling" to the previous category or as a "child" to the previous
				category is determined by the number of spaces that prefix the line.
*********************************************************************************************/
void Category::fill_categories(const vector<string> & lines, int & current_line)
{
    int last_line_added = 0;
	string categoryname;
	int end_of_categoryname;
	int lead_spaces_current, lead_spaces_previous;
	
	
	if (current_line < lines.size())
	{
	    //truncate "line" to get category name
		end_of_categoryname = lines[current_line].find_first_of("\t\r");
      	if (end_of_categoryname == string::npos)
    	{
    		end_of_categoryname = lines[current_line].length();
    	}
      	const int start_of_categoryname = lines[current_line].find_first_not_of (" ");
		categoryname = lines[current_line].substr(start_of_categoryname, end_of_categoryname - start_of_categoryname);
   		
   		//push_back current "subcategories" vector to add subactegory
   		subcategories.push_back(Category(categoryname));
   		
   		//add files if this subcategory contains files
   		if (lines[current_line].find("\t")!= string::npos)
   		{
   		    subcategories.back().files = get_filenames(lines[current_line]);
   		}
   		
        //keep subscript of last line added to Category in order to compare with next line
   		last_line_added = current_line++;
  }


  while (current_line < lines.size() )
  {
        //truncate "line" to get category name
        end_of_categoryname = lines[current_line].find_first_of("\t\r");
      	if (end_of_categoryname == string::npos)
    	{
    		end_of_categoryname = lines[current_line].length();
    	}
      	const int start_of_categoryname = lines[current_line].find_first_not_of (" ");
       	categoryname = lines[current_line].substr(start_of_categoryname, end_of_categoryname - start_of_categoryname);
        		  			     			
       	
       	lead_spaces_previous = lines[last_line_added].find_first_not_of(" ");
       	lead_spaces_current = lines[current_line].find_first_not_of(" ");
		    	          		      	 		
       	//if current line is a child of the previous, call function recursively to add line as a subcategory
       	if (lead_spaces_current > lead_spaces_previous)
       	{
       		subcategories.back().fill_categories(lines, current_line);
       	}
       	//if current line is a sibling of the previous, push_back the current subcategories vector
       	else if (lead_spaces_current == lead_spaces_previous)
      	{	      		
      		subcategories.push_back(Category(categoryname));
      		if (lines[current_line].find("\t")!= string::npos)
   			{
   		    	subcategories.back().files = get_filenames(lines[current_line]);
   			}
   			
      		last_line_added = current_line++;
       	}
       	
       	//if current line is not a child or a sibling of previous,  return so that it can be added
       	//to the appropriate subcategory vector
       	else
       	{       	 		
        		return;
       	}	       		  	 	
         		
    } //end while loop	    			
}	

/********************************************************************************************
	Function: Category::find_id
	Receives: const wxTreeItemId &selected_id
	Returns: const Category *
	Description: Finds and returns a pointer to the constant Category object associated with
				the given wxTreeItemId.
*********************************************************************************************/
const Category * Category::find_id (const wxTreeItemId & selected_id) const
{
	if (selected_id == category_tree_id)
	{
		return this;
	}
	else
	{
		const Category * node = NULL;
		
		for (int i = 0; i < subcategories.size(); i++)
		{
			if ((node = subcategories[i].find_id (selected_id)) != NULL)
			{
				return node;
			}
		}
		
		return node;
	}
}

/********************************************************************************************
	Function: Category::find_id
	Receives: const wxTreeItemId &selected_id
	Returns: const Category *
	Description: Finds and returns a pointer to the non-constant Category object associated with
				the given wxTreeItemId.
*********************************************************************************************/
Category * Category::find_id (const wxTreeItemId & selected_id)
{
	if (selected_id == category_tree_id)
	{
		return this;
	}
	else
	{
		Category * node = NULL;
		
		for (int i = 0; i < subcategories.size(); i++)
		{
			if ((node = subcategories[i].find_id (selected_id)) != NULL)
			{
				return node;
			}
		}
		
		return node;
	}
}


/********************************************************************************************
	Function: Category::format_category_file
	Receives: vector string & lines, string spaces
	Returns: void
	Description: Fills a vector of strings, each element representing one line in a text file
				of categories, specially formatted to be recursively into a family of
				category objects.
*********************************************************************************************/

void Category::format_category_file(vector <string> &lines, string spaces)
{
	string line = spaces + name;
	
	for (int i = 0; i < files.size(); i++)
	{
		line+= "\t" + files[i];
	}
	
	lines.push_back(line);
	
	spaces+= " ";
	for (int i = 0; i < subcategories.size(); i++)
	{
		subcategories[i].format_category_file(lines, spaces);
		
	}
	
}

/********************************************************************************************
	Function: Category::get_filenames
	Receives: const string & line
	Returns: const vector<string>
	Description: Receives a string that consists of one or more filenames separated by
				<TAB>spaces. Splits the string into individual filenames fills a vector with
				the resulting token strings.
*********************************************************************************************/
const vector<string> Category::get_filenames(const string & line)
{
    int start_pos = 0, end_pos = 0;
    string file_name;
    vector <string> files;

    do
    {
    	start_pos = line.find_first_of("\t\r", start_pos) + 1;
    	 	
    	if (line.find_first_of("\t\r", start_pos)!= string::npos)
    		end_pos = line.find_first_of("\t\r", start_pos);
    	else
    		end_pos = line.length();
    		
    	file_name = line.substr(start_pos, end_pos - start_pos);
    	
    	if (file_name.length() > 0  )
    		files.push_back(file_name);
    		
    	start_pos += file_name.length();
    	    	
    }   while (start_pos < line.length() );
	
	return files;

}	

/********************************************************************************************
	Function: Category::get_files
	Receives: void
	Returns: const vector<string> &
	Description: Returns a constant reference to the "files" data member of a Category object.
*********************************************************************************************/
const vector<string> & Category::get_files() const
{
	return files;
}


/********************************************************************************************
	Function: Category::get_files
	Receives: void
	Returns: const vector<string> &
	Description: Returns a constant reference to the "files" data member of the Category object
				associated with the given wxTreeItemId.
*********************************************************************************************/

const vector<string> & Category::get_files(const wxTreeItemId & selected_id) const
{
	static vector<string> error;
	
	const Category * node = find_id (selected_id);
	
	if (node != NULL)
	{
		return node->get_files();
	}
	else
	{
		return error;
	}
}


/********************************************************************************************
	Function: Category::get_name
	Receives: void
	Returns: const string &
	Description: Returns a constant reference to the "name" data member of a Category object.
*********************************************************************************************/
const string & Category::get_name() const
{
	return name;
}

/********************************************************************************************
	Function: Category::get_name
	Receives: void
	Returns: const string &
	Description: Returns a constant reference to the "name" data member of the Category object
				associated with the given wxTreeItemId.
*********************************************************************************************/

const string & Category::get_name(const wxTreeItemId & selected_id)const
{
	static string empty = "";
	const Category * node = find_id (selected_id);
	
	if (node != NULL)
	{
		return node->name;		
	}	
	else
	{
		return empty;
	}	
}

/********************************************************************************************
	Function: Category::get_subcategories
	Receives: void
	Returns: const vector<Category> &
	Description: Returns a constant reference to the "subcategories" data member of a Category object.
*********************************************************************************************/
const vector<Category> & Category::get_subcategories() const
{
 	return subcategories;

}

/********************************************************************************************
	Function: Category::get_subcategories
	Receives: void
	Returns: const vector<Category> &
	Description: Returns a non-constant reference to the "subcategories" data member of a Category object.
*********************************************************************************************/
vector<Category> & Category::get_subcategories()
{
 	return subcategories;

}


/********************************************************************************************
	Function: Category::set_tree_id
	Receives: const wxTreeItemId & selected_id
	Returns:  void
	Description: Assigns a value to the wxTreeItemId data member of a Category object.
*********************************************************************************************/

void Category::set_tree_id(const wxTreeItemId & id)
{
	category_tree_id = id;	
}


/********************************************************************************************
	Function: Category::modify_category
	Receives: const string &new_name, const wxTreeItemId & selected_id
	Returns:  void
	Description: Assigns a new_name to the Category object associated with the given wxTreeItemId.
*********************************************************************************************/

void Category::modify_category(const string & new_name, const wxTreeItemId & selected_id)
{	
	Category * node = find_id (selected_id);
	
	if (node != NULL)
	{
		node->name = new_name;
	}	
}


/********************************************************************************************
	Function: Category::remove
	Receives: const string &filename, const wxTreeItemId & selected_id
	Returns:  void
	Description: Erases one one element of the "files" vector in the Category object associated
				with the given wxTreeItemId.
*********************************************************************************************/

void Category::remove_file(const string &file_name, const wxTreeItemId & selected_id)
{

	Category * node = find_id(selected_id);
	
	if (node != NULL)
	{
		vector <string>::iterator i;
		for (i = node->files.begin(); i != node->files.end(); ++i)
		{
			if ( *i == file_name )
			{
				i = node->files.erase (i);
				return;				
			}
		}
	}
		
}

/********************************************************************************************
	Function: Category::remove
	Receives: wxTreeItemId & parent_id, const wxTreeItemId & selected_id
	Returns:  void
	Description: removes the element of the subcategories vector which "owns" the Category
				object/wxTreeItemId selected by the user.
*********************************************************************************************/

void Category::remove_subcategory(const wxTreeItemId & parent_id, const wxTreeItemId & selected_id)
{
	Category * node = find_id (parent_id);
		
	if ( node != NULL)
	{
		vector <Category>::iterator i;
		for (i = node->subcategories.begin(); i != node->subcategories.end(); ++i)
		{
			if ((*i).category_tree_id == selected_id)
			{
				i = node->subcategories.erase (i);
				return;
			}
		}
	}	
}
