/**********************************************************************************************************

Programmers:			Muhammad Zulfigar, Thalia Lewis
filename:: 				category_treectrl.h
						Includes functions pertaining to class CategoryTreeCtrl
Date:					August 8, 2000

***********************************************************************************************************/

#include "category_treectrl.h"

#include "category.h"

#include "wx/log.h"
#include "wx/image.h"
#include "wx/treectrl.h"
#include <wx/string.h>
#include "wx/treectrl.h"
#include "wx/msgdlg.h"
#include "wx/utils.h"
#include "wx-view.h"

#include <string>
#include <vector>
#include <fstream>
using namespace std;

/********************************************************************************************
	Function: CategoryTreeCtrl::CategoryTreeCtrl
	Receives: N/A (constructor)
	Returns: N/A (constructor)
	Description:
*********************************************************************************************/
CategoryTreeCtrl::CategoryTreeCtrl(wxWindow *parent, const wxWindowID id)
	: wxTreeCtrl(parent, id, wxDefaultPosition, wxDefaultSize, wxSUNKEN_BORDER),
	  last_click_position (10, 10)
{

	vector<string> lines;
	
	lines = get_category_file_contents();   //fill vector with contents of "categories.txt"
	
	if (!lines.empty())
	{
		int current_line = 0;
		root_category_ptr = new Category(lines[current_line]); //initialize "root" Category object
		root_category_ptr->fill_categories(lines, ++current_line); //recursively fill "root" Category
																	//with contents of "categories.txt"
		
		//construct the treectrl
		root_id = AddRoot(root_category_ptr->get_name().c_str(), -1, -1);
    	root_category_ptr->set_tree_id(root_id);
    	build_tree(root_id, *root_category_ptr);


		//create pop-up menu
		menu = new wxMenu;        	
		menu->Append (ID_add_subcategory, translation["Add Subcategory"]);
		menu->AppendSeparator();
		menu->Append (ID_modify_category, translation["Modify Category"]);
		menu->AppendSeparator();
		menu->Append (ID_remove_category, translation["Remove Category"]);
   }
}

/********************************************************************************************
	Function: CategoryTreeCtrl::EVENT_TABLE
	Receives: void
	Returns: void
	Description: Lists treectrl events, or events occurring on treectrl pop-menu
*********************************************************************************************/
BEGIN_EVENT_TABLE(CategoryTreeCtrl, wxTreeCtrl)
	EVT_RIGHT_UP (CategoryTreeCtrl::OnRightClick)
	EVT_LEFT_UP (CategoryTreeCtrl::OnLeftClick)
	EVT_MENU (ID_add_subcategory, CategoryTreeCtrl::OnMenuAddCategory)
	EVT_MENU (ID_remove_category, CategoryTreeCtrl::OnMenuRemoveCategory)
	EVT_MENU (ID_modify_category, CategoryTreeCtrl::OnMenuModifyCategory)
END_EVENT_TABLE()


/********************************************************************************************
	Function: CategoryTreeCtrl::add_category ()
	Receives: wxWindow * parent, const wxTreeItemId & selected_id
	Returns: void
	Description: Allows the user to add a sub-category to the Category object associated
				with the given wxTreeItemId. Displays a text entry box on the parent window.
*********************************************************************************************/
void CategoryTreeCtrl::add_category(wxWindow * parent, const wxTreeItemId & selected_id)
{

   	wxTextEntryDialog new_category(parent, translation["Add Category"], translation["Enter Category Name: "], "", wxOK | wxCANCEL | wxCENTRE, wxDefaultPosition);
    new_category.ShowModal();

    //get the name of the new category
    wxString name = new_category.GetValue();
    if (name.Length() == 0)
    	return;	

	//add new Category object
	string new_category_name = name.c_str();
	root_category_ptr->add_subcategory(new_category_name, selected_id);
 	update_category_file_contents();
  	
  	//add Category to treectrl
  	wxTreeItemId new_id = AppendItem(selected_id, new_category_name.c_str(), -1, -1);
  	add_new_id(selected_id, new_id);
}	


/********************************************************************************************
	Function: CategoryTreeCtrl::add_file () const
	Receives: const wxTreeItemId & selected_id
	Returns: void
	Description: Allows the user to add a file to the Category object associated with the
				given wxTreeItemId.
*********************************************************************************************/
bool CategoryTreeCtrl::add_file(wxWindow * parent, const wxString & file_name, const wxTreeItemId & selected_id) const
{
 	wxString message = translation["Cannot add files to root category"];
 	if ( selected_id == root_id )
	{
	 	wxMessageDialog root_category(parent, message, translation["Root Category"], wxOK | wxICON_EXCLAMATION, wxDefaultPosition);
		root_category.ShowModal();
		return false;
	}
 	root_category_ptr->add_file(file_name.c_str(), selected_id);
 	update_category_file_contents();
 	return true;
}



/********************************************************************************************
	Function: CategoryTreeCtrl::add_new_id
	Receives: const wxTreeItemId & selected_id, const wxTreeItemId & new_id
	Returns: void
	Description: Assigns a wxTreeItemId to the Category object associated with the item that has
				just been added to the treectrl.
*********************************************************************************************/
void CategoryTreeCtrl::add_new_id(const wxTreeItemId & selected_id, const wxTreeItemId & new_id)
{

	Category * parent_category = get_category(selected_id);
	if (parent_category == NULL)
		return;
		
	parent_category->get_subcategories().back().set_tree_id(new_id);	
	
}


/********************************************************************************************
	Function: CategoryTreeCtrl::build_tree()
	Receives: const wxTreeItemId & parent_id, Category & category, bool first_entry = true
	Returns: void
	Description: Function recursively builds an entire CategoryTreeCtrl object, and
				simultaneously assigns a wxTreeItemId to every category object that
				represents a branch on the tree.
*********************************************************************************************/
void CategoryTreeCtrl::build_tree(const wxTreeItemId &parent_id, Category & category, bool first_entry)
{ 	
 	wxTreeItemId child_id;
 	wxString categoryname;
 	categoryname =  category.get_name().c_str(); 	
 	
 	//if entering function for the first time, treectrl does not add the "root" Category as a node on the tree
 	//since it has already been added as the root of the treectrl (see Constructor)
 	
 	if (first_entry == false)
 	{
 	    child_id = AppendItem(parent_id, categoryname, -1, -1);
 	    category.set_tree_id(child_id);
 	}
 	else  //if first_entry == true
 	{
 	    child_id = parent_id;
 	    first_entry = false;
 	}
 	
 	if (!category.get_subcategories().empty())
 	{
 		for (int i = 0; i < category.get_subcategories().size(); i++)
 		     build_tree(child_id, category.get_subcategories()[i], first_entry); 			
 	}
 	 	
}


/********************************************************************************************
	Function: CategoryTreeCtrl::get_category () const
	Receives: const wxTreeItemId & selected_id
	Returns: const Category *
	Description: Returns a pointer to the constant Category object associated with the
				given wxTreeItemId.
*********************************************************************************************/
const Category * CategoryTreeCtrl::get_category(const wxTreeItemId & selected_id) const
{
	Category * found_category = root_category_ptr->find_id(selected_id);
	return found_category;
}



/********************************************************************************************
	Function: CategoryTreeCtrl::get_category ()
	Receives: const wxTreeItemId & selected_id
	Returns: Category *
	Description: Returns a pointer to the non-constant Category object associated with
				the given wxTreeItemId.
*********************************************************************************************/
Category * CategoryTreeCtrl::get_category(const wxTreeItemId & selected_id)
{
	Category * found_category = root_category_ptr->find_id(selected_id);
	return found_category;
}




/********************************************************************************************
	Function: CategoryTreeCtrl::get_category_file_contents()const
	Receives: void
	Returns: const vector <string> lines
	Description: Returns a vector of strings representing each line in a text file of
				categories, sub-categories and filenames.
*********************************************************************************************/
const vector<string> CategoryTreeCtrl::get_category_file_contents() const
{
	vector<string> lines;
	string line;
    //get full path to directory that contains the text file containing data on user's categories
	string full_path = user_preferences.fullpath + "/categories.txt";
	
	ifstream file_ptr(full_path.c_str());  //text file where categories stored
	
	if (file_ptr)
	{	
  		while (getline(file_ptr, line))	 	 	
     	{	
     		if(line != "")
     			lines.push_back(line);
     	}	
	}
	else
	{
	    lines.push_back (translation["Categories"]);
	}
	
	return lines;
	
}


/********************************************************************************************
	Function: CategoryTreeCtrl::get_filenames
	Receives: vector<wxString> &file_names, const wxTreeItemId & selected_id
	Returns: void
	Description: Returns the vector of filenames belonging to the Category object
				associated with the given wxTreeItemId.
*********************************************************************************************/
void CategoryTreeCtrl::get_filenames(vector<wxString> &file_names, const wxTreeItemId & selected_id) const
{
 	vector <string> files;
	
	files = root_category_ptr->get_files(selected_id);
	
	for (int i = 0; i < files.size(); i++)
	{
	 	file_names.push_back(files[i].c_str());
	}
	
}


/********************************************************************************************
	Function: CategoryTreeCtrl::get_root_category
	Receives: void
	Returns: const Category *
	Description: Returns a pointer to the constant Category Object containing all items in the
				 CategoryTreeCtrl
*********************************************************************************************/
const Category * CategoryTreeCtrl::get_root_category()const
{
 	return root_category_ptr;
}


/********************************************************************************************
	Function: CategoryTreeCtrl::get_root_category()
	Receives: void
	Returns: Category *
	Description: Returns a pointer to the Category Object containing all items in the
				 CategoryTreeCtrl
*********************************************************************************************/
Category * CategoryTreeCtrl::get_root_category()
{
 	return root_category_ptr;
}




/********************************************************************************************
	Function: CategoryTreeCtrl::modify_category
	Receives: wxWindow * parent, const wxTreeItemId & selected_id
	Returns: void
	Description: Allows user to modify the name of the Category Object associated with the
				given wxTreeItemId.
*********************************************************************************************/
void CategoryTreeCtrl::modify_category(wxWindow * parent, const wxTreeItemId & selected_id)	
{
	string category_name = root_category_ptr->get_name(selected_id);
	if (category_name.length() == 0)  //if cannot find Category associated with selected ID, quit function
		return;
		
	wxTextEntryDialog rename_category(parent, translation["Modify Category"], translation["Enter Category Name: "], category_name.c_str(), wxOK | wxCANCEL | wxCENTRE, wxDefaultPosition);
    rename_category.ShowModal();
    string new_name = rename_category.GetValue().c_str();
    if (new_name.length() != 0)
    {
    	root_category_ptr->modify_category(new_name, selected_id);
    	update_category_file_contents();
    	SetItemText (selected_id, new_name.c_str());
    }
}



/********************************************************************************************
	Function: CategoryTreeCtrl::remove_category ()
	Receives: wxWindow * parent, const wxTreeItemId & selected_id
	Returns: void
	Description: Allows the user to remove the category object(and its sub-categories, if any)
				that is associated with the given wxTreeItemId. Displays a message box on the
				parent window to the CategoryTreeCtrl object. Does not permit the user to remove
				the "root" category.
*********************************************************************************************/
void CategoryTreeCtrl::remove_category(wxWindow * parent, const wxTreeItemId & selected_id)
{
	string name = root_category_ptr->get_name(selected_id);
	if (name.length() == 0) //if cannot find Category associated with selected ID, quit function
		return;
	
	
	wxString category_name = name.c_str();
	wxString message = translation["Cannot delete '"] + category_name + "'";
	//User CANNOT delete the "root" Category
	if ( selected_id == root_id )
	{
	 	wxMessageDialog root_category(parent, message, translation["Root Category"], wxOK | wxICON_EXCLAMATION, wxDefaultPosition);
		root_category.ShowModal();
		return;
	}
	
	message = translation["Delete: '"] + category_name + "'";
	wxMessageDialog confirm_remove_category(parent, message, translation["Delete Category"], wxYES_NO | wxICON_QUESTION, wxDefaultPosition);
	
	if (confirm_remove_category.ShowModal() != wxID_NO )
	{
		wxTreeItemId parent_id =  GetParent(selected_id);
		root_category_ptr->remove_subcategory(parent_id, selected_id);
 		Delete(selected_id);
 		update_category_file_contents();
 		SelectItem(parent_id); //highlights parent_Category after Category has been deleted
 	}
}

/********************************************************************************************
	Function: CategoryTreeCtrl::remove_file () const
	Receives: const wxTreeItemId & selected_id
	Returns: void
	Description: Allows the user to delete a file from the category object that is associated
				with the given wxTreeItemId.Displays a message box on the parent window to the
				CategoryTreeCtrl object.
*********************************************************************************************/
void CategoryTreeCtrl::remove_file(wxWindow * parent, const wxString & file_name) const
{
 	
 	wxTreeItemId selected_id = GetSelection();
 	
 	//truncate file_name
 	wxString file = file_name;
 	file = file.AfterLast('/');
 	file = file.BeforeFirst('.');
 	
 	wxString message = translation["Remove: '"] + file + "'";
	wxMessageDialog confirm_remove_file(parent, message, translation["Remove Image From Category"], wxYES_NO | wxICON_QUESTION, wxDefaultPosition);
	
 	if (confirm_remove_file.ShowModal() != wxID_NO )
	{ 	
 		root_category_ptr->remove_file(file_name.c_str(), selected_id);
 		update_category_file_contents(); 		
 	}
 	
}


/********************************************************************************************
	Function: CategoryTreeCtrl::update_category_file_contents
	Receives: void
	Returns: void
	Description: Writes information from Category objects contained in CategoryTreeCtrl
				 to "categories" text file.		
*********************************************************************************************/

void CategoryTreeCtrl::update_category_file_contents() const
{
	vector <string> lines;
    root_category_ptr->format_category_file(lines, "");
	
	string full_path = user_preferences.fullpath + "/categories.txt";

    ofstream updated_file(full_path.c_str());

    if (updated_file)
	{	
  		for (int i = 0; i < lines.size(); i++)
  		{
  			updated_file << lines[i] << endl;
  		}
	}	
}
	


//EVENTS
/********************************************************************************************
	Function: CategoryTreeCtrl::OnRightClick
	Receives: wxMouseEvent& event
	Returns: void
	Description: When user right-clicks on treectrl, pop-up menu to add, modify, or remove a
				category appears.		
*********************************************************************************************/

void CategoryTreeCtrl::OnRightClick(wxMouseEvent & event)
{
	if (last_click_position != wxPoint(-1,-1))
	{        	
 		PopupMenu (menu, last_click_position);
	}	
}

/********************************************************************************************
	Function: CategoryTreeCtrl::OnLeftClick
	Receives: wxMouseEvent &event
	Returns: void
	Description: When user selects an item on the treectrl, the coordinates of the position of
				the left-click are stored so that the pop-up menu appears in the correct locateion. 		
*********************************************************************************************/
void CategoryTreeCtrl::OnLeftClick (wxMouseEvent & event)
{
   	last_click_position = event.GetPosition();
   	event.Skip();
}



/********************************************************************************************
	Function: CategoryTreeCtrl::OnMenuAddCategory
	Receives: wxMouseEvent& event
	Returns: void
	Description: If user selects this item on pop-up menu, dialog appears to add a subcategory
				to the Category associated with the currently-selected item on treectrl.	
*********************************************************************************************/
void CategoryTreeCtrl::OnMenuAddCategory (wxCommandEvent &)
{
	add_category(this, GetSelection());	
}


/********************************************************************************************
	Function: CategoryTreeCtrl::OnMenuAddCategory
	Receives: wxMouseEvent& event
	Returns: void
	Description: If user selects this item on pop-up menu, dialog appears to remove the Category
				associated with currently-selected item on treectrl.	
*********************************************************************************************/
void CategoryTreeCtrl::OnMenuRemoveCategory (wxCommandEvent &)
{
	remove_category(this, GetSelection());
}

/********************************************************************************************
	Function: CategoryTreeCtrl::OnMenuAddCategory
	Receives: wxMouseEvent& event
	Returns: void
	Description: If user selects this item on pop-up menu, dialog appears to modify the Category
				associated with currently-selected item on treectrl.	
*********************************************************************************************/
void CategoryTreeCtrl::OnMenuModifyCategory (wxCommandEvent &)
{
    modify_category(this, GetSelection());
}

