/******************************************
imagepanel.cpp
Mike MacIntyre, David Boys
Function definitions for Class ImagePanel
last updated : August 9, 2000
******************************************/

#include "imagepanel.h"

#include "PreviewPanel.h"
#include "imagedialog.h"

#include "wx/file.h"

BEGIN_EVENT_TABLE(ImagePanel, wxPanel)
	EVT_LEFT_UP(ImagePanel::OnLeftClick)
	EVT_LEFT_DCLICK(ImagePanel::OnLeftDClick)
	EVT_TIMER (ID_timer, ImagePanel::OnDClickTimeout)
END_EVENT_TABLE()

///// FUNCTION DEFINITIONS /////

// *** ImagePanel Class ***

/**********************************************************************
   Function: ImagePanel::ImagePanel
   Receives: wxWindow *, const wxWindowID, const wxPoint&, const wxSize&, long,
   			 const wxString &, const int, vector <wxString> &, vector<ImagePanel *> &,
   			 const wxImage *, PreviewPanel *
   Returns:  N/A (Constructor)
   Description: Constructs a custom panel with an image icon, and a label with
   				the file name on the label
**********************************************************************/
ImagePanel::ImagePanel(wxWindow * parent, const wxWindowID id, const wxPoint& pos,
    	 	           const wxSize& size, long style , const wxString & path,
                	   const int config_icon_size, vector <wxString> & temp,
                	   vector<ImagePanel *> & panels, const wxImage * img,
            	 	   PreviewPanel * preview)
    	 	
	 	  : wxPanel(parent, id, pos, size, style), full_path(path), temp_selections(temp),
	 	    panels(panels), preview(preview), dbl_click_timer(this, ID_timer)
	 	
{
	int h = config_icon_size;
	int bitmap_Y;
	
    SetBackgroundColour(* wxWHITE);

    activated = false;

	file_name_label = new clickableStaticText(this, -1, "", wxPoint(0, 0), wxSize(0, 0));

	file_name_label -> SetLabel(wxFileNameFromPath(full_path));	
	file_name_label -> Move(config_icon_size + 15, (h - label_height())/ 2);
	
	SetSize(config_icon_size + label_width() + 20, h);
	

    image = new wxImage (*img);
    bitmap_Y = (h - image->GetHeight()) / 2;

    bm = new clickableStaticBitmap (this, -1, image->ConvertToBitmap(), wxPoint(0, bitmap_Y));
}

/**********************************************************************
   Function: ImagePanel::~ImagePanel
   Receives: N/A (destructors)
   Returns:  N/A (destructors)
   Description:  Deallocates memory.
**********************************************************************/
ImagePanel::~ImagePanel()
{
	delete (image);
	delete (bm);	
}

/**********************************************************************
   Function: ImagePanel::reset
   Receives: N/A, no parameters
   Returns:  void
   Description: Resets the panel to the non-highlighted state
**********************************************************************/
void ImagePanel::reset()
{
	file_name_label->SetForegroundColour(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_WINDOWTEXT));
	SetBackgroundColour(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_WINDOW));
	Refresh();
	activated = false;
}

/**********************************************************************
   Function: ImagePanel::OnLeftClick
   Receives: wxMouseEvent &
   Returns:  void
   Description: Captures the left single click, the left double click,
   				the shift left click, and the ctlr-left click events
   				on the panels, image icons, and static texts.   				
**********************************************************************/
void ImagePanel::OnLeftClick(wxMouseEvent & event)
{
    const bool normal_click = (!event.m_controlDown) && (!event.m_shiftDown) && (!event.m_altDown);
    const bool ctrl_click = event.m_controlDown && (!event.m_shiftDown) && (!event.m_altDown);
    const bool shift_click = event.m_shiftDown && (!event.m_controlDown) && (!event.m_altDown);


    ImagePanel * current_selection = this;

    	// Avoid repeating the image preview processing
	if ((normal_click || shift_click) && activated && temp_selections.size() == 1)
	{
		return;
	}
	
	if (shift_click && dynamic_cast <FileCanvas *> (GetParent())->get_previous() != NULL)
	{
	    clear_preview_image();
	    temp_selections.clear();
	    deselect_all();
	
	    int prev, cur;
	
		for (int i = 0; i < panels.size(); i++)
		{
		    if (dynamic_cast <FileCanvas *> (GetParent())->get_previous() == panels[i])
		        prev = i;
		
		    if (current_selection == panels[i])
		        cur = i;
		}
		
		for (int i = min(prev, cur); i <= max(prev, cur); i++)
		{
        	panels[i] -> select_panel();
            temp_selections.push_back(panels[i] -> full_path);		
		}
	
	}
    else
    {
    	if (ctrl_click && activated)	// ctrl-click on an active item -- deselect it and remove if from
	    {
		    reset();
    		vector<wxString>::iterator pos = find(temp_selections.begin(), temp_selections.end(), full_path);
		
	    	if (pos != temp_selections.end())
		    {
		        temp_selections.erase (pos);
    		}
	    }
    	else
	    {
	        clear_preview_image();
	
        	if (normal_click)
        	{
    	    	temp_selections.clear();
    		    deselect_all();	
        	}

        	select_panel();
            temp_selections.push_back(full_path);		
            dynamic_cast <FileCanvas *> (GetParent())->set_previous(this);
        }

    }

	if (temp_selections.size() == 1)
	{
		const bool ONE_SHOT = true;
		dbl_click_timer.Start (DCLICK_TIMEOUT, ONE_SHOT);
	}
}

/**********************************************************************
   Function: ImagePanel::select_panel
   Receives: N/A (no-parameters)
   Returns:  void
   Description: Sets the panel to a highlighted state
**********************************************************************/
void ImagePanel::select_panel()
{
   	file_name_label -> SetForegroundColour(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_HIGHLIGHTTEXT));
   	SetBackgroundColour(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_HIGHLIGHT));
   	Refresh();
    activated = true;
}

/**********************************************************************
   Function: ImagePanel::OnLeftDClick
   Receives: wxMouseEvent &
   Returns:  void
   Description: Processes the left double click event, calling the
   				process_ok function to close the dialog box.
**********************************************************************/
void ImagePanel::OnLeftDClick (wxMouseEvent & event)
{
    dbl_click_timer.Stop();
    dynamic_cast<ImageDialog *>(GetGrandParent())->process_ok();
}

/**********************************************************************
   Function: ImagePanel::OnDClickTimeout
   Receives: wxTimerEvent &
   Returns:  void
   Description: After the timer event has elapsed, call the load the
   				load_preview_image function
**********************************************************************/
void ImagePanel::OnDClickTimeout (wxTimerEvent &)
{
        // Timeout ellapsed: No double-click, complete the single-click actions
    load_preview_image();
}

/**********************************************************************
   Function: ImagePanel::load_preview_image
   Receives: N/A (no parameters)
   Returns:  void
   Description: Loads the preview image directly from the file and
   				places it on preview panel; passes the parameters
   				to display the image information on the
   				dialog box(ImageDialog)
**********************************************************************/
void ImagePanel::load_preview_image()
{
    if (temp_selections.size() == 1)
    {
        wxFile f (temp_selections[0]);  //(full_path);  (BUG FIX -- D.B. and C.M. -- when ctrl-clicking and deselcting one, the
                                                      // one deselected is the one that was showing)

    	off_t bytes = f.Length();
    	
    	wxBeginBusyCursor();
    	wxYield();
        image = new wxImage();
    	image->LoadFile (temp_selections[0], wxBITMAP_TYPE_ANY);  //(full_path, wxBITMAP_TYPE_ANY);  (BUG FIX -- see above)
    	dynamic_cast<ImageDialog *>(GetGrandParent())-> show_preview_info(image->GetHeight(),
    									image->GetWidth(), static_cast <unsigned long> (bytes));	
    	preview->preview_image(image);

    	wxEndBusyCursor();
	}
}

/**********************************************************************
   Function: ImagePanel::clear_preview_image
   Receives: N/A (no parameters)
   Returns:  void
   Description: Clears the preview image from the preview panel and clears
   				the image information from the dialog box(ImageDialog)
**********************************************************************/
void ImagePanel::clear_preview_image()
{
	preview->clear_preview();
	dynamic_cast<ImageDialog *>(GetGrandParent())-> hide_preview_info();	
	
}

/**********************************************************************
   Function: ImagePanel::deselect_all
   Receives: N/A (no parameters)
   Returns:  void
   Description: Deselects all active/highlighted panels
**********************************************************************/
void ImagePanel::deselect_all()
{
	for(int i = 0; i < panels.size(); i++)
	{
		if (panels[i]->activated == true)
		{
	    	panels[i]->reset();
	    	panels[i]->activated = false;
	 	}
	}
}		

/**********************************************************************
   Function: ImagePanel::label_width
   Receives: N/A (no parameters)
   Returns:  int
   Description: Returns the label width
**********************************************************************/
int ImagePanel::label_width() const
{
	int w, h;
	file_name_label -> GetSize(&w, &h);
	return w;
}

/**********************************************************************
   Function: ImagePanel::label_height
   Receives: N/A (no parameters)
   Returns:  int
   Description: Returns the label height
**********************************************************************/
int ImagePanel::label_height() const
{
	int w, h;
	file_name_label -> GetSize(&w, &h);
	return h;
}


// *** ClickableStaticBitmap Class ***

/**********************************************************************
   Function: clickableStaticBitmap::clickableStaticBitmap
   Receives: ImagePanel *, const wxWindowID, const wxBitmap &,
   			 const wxPoint &, const wxSize&
   Returns:  N/A (Constructor)
   Description: Creates a derived class of a wxBitmap, to catch the
   				click events
**********************************************************************/
clickableStaticBitmap::clickableStaticBitmap (ImagePanel * parent, const wxWindowID id, const wxBitmap & bmp, const wxPoint & pos, const wxSize & size)
        : wxStaticBitmap (parent, id, bmp, pos, size)
{
    Connect (-1, wxEVT_LEFT_UP, reinterpret_cast<wxObjectEventFunction>(&clickableStaticBitmap::OnClick));
    Connect (-1, wxEVT_LEFT_DCLICK, reinterpret_cast<wxObjectEventFunction>(&clickableStaticBitmap::OnDblClick));
}

/**********************************************************************
   Function: clickableStaticBitmap::OnClick
   Receives: wxMouseEvent &
   Returns:  void
   Description: Catches the left click event and passes it to the
   				parent (ImagePanel) OnLeftClick event.
**********************************************************************/
void clickableStaticBitmap::OnClick (wxMouseEvent & event)
{
	dynamic_cast<ImagePanel *>(GetParent())->OnLeftClick(event);
}

/**********************************************************************
   Function: clickableStaticBitmap::OnDblClick
   Receives: wxMouseEvent &
   Returns:  void
   Description: Catches the left double click event and passes it to
   				the parent (ImagePanel) OnLeftDClick event.
**********************************************************************/
void clickableStaticBitmap::OnDblClick (wxMouseEvent & event)
{
	dynamic_cast<ImagePanel *>(GetParent())->OnLeftDClick(event);
}


// *** ClickableStaticText Class ***

/**********************************************************************
   Function: clickableStaticText::clickableStaticText
   Receives: ImagePanel *, const wxWindowID, const wxString &,
   			 const wxPoint &, const wxSize &
   Returns:  N/A (Constructor)
   Description: Creates a derived class of a wxStaticText, to catch the
   				click events
**********************************************************************/
clickableStaticText::clickableStaticText (ImagePanel * parent, const wxWindowID id, const wxString & txt, const wxPoint & pos, const wxSize & size)
        : wxStaticText (parent, id, txt, pos, size, wxALIGN_LEFT)
{
    Connect (-1, wxEVT_LEFT_UP, reinterpret_cast<wxObjectEventFunction>(&clickableStaticText::OnClick));
    Connect (-1, wxEVT_LEFT_DCLICK, reinterpret_cast<wxObjectEventFunction>(&clickableStaticText::OnDblClick));
}

/**********************************************************************
   Function: clickableStaticText::OnClick
   Receives: wxMouseEvent &
   Returns:  void
   Description: Catches the left click event and passes it to the
   				parent (ImagePanel) OnLeftClick event.
**********************************************************************/
void clickableStaticText::OnClick (wxMouseEvent & event)
{
	dynamic_cast<ImagePanel *>(GetParent())->OnLeftClick(event);
}

/**********************************************************************
   Function: clickableStaticText::OnDblClick
   Receives: wxMouseEvent &
   Returns:  void
   Description: Catches the left double click event and passes it to
   				the parent (ImagePanel) OnLeftDClick event.
**********************************************************************/
void clickableStaticText::OnDblClick (wxMouseEvent & event)
{
	dynamic_cast<ImagePanel *>(GetParent())->OnLeftDClick(event);
}
